//+------------------------------------------------------------------+
//|                                           CustPrevCalculated.mqh |
//|                                         Copyright 2011, AlexSTAL |
//|                                           http://www.alexstal.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, AlexSTAL"
#property link      "http://www.alexstal.ru"

//+------------------------------------------------------------------+
//| Result of working of the function of history synchronization     |
//+------------------------------------------------------------------+
enum CPCPrepareDataResultCode
  {
   CPCPDRC_NoData,                     // Returned when there is no data for calculation (not [prepared by server) or the class is not initialized
   CPCPDRC_FullInitialization,         // Full initialization of the array has been performed
   CPCPDRC_Synch,                      // Synchronization of adding of new bars has been performed
   CPCPDRC_SynchOnlyLastBar,           // Synchronization of the last bar has been performed (with probable cutting of history)
   CPCPDRC_NoRecountNotRequired        // No recalculation since the data hasn't been changed
  };

//+------------------------------------------------------------------+
//| Mode of working of the history synchronization                   |
//| Mode depth depends on HistorySynchSecond                         |
//+------------------------------------------------------------------+
enum CPCHistorySynchMode
  {
   CPCHSM_NotSynch,                    // No synchronization
   CPCHSM_Normal,                      // Mode of working if the number of bars is changed
   CPCHSM_Paranoid                     // Paranoid mode of working at every tick
  };

//+------------------------------------------------------------------+
//| Auxiliary: the result of working of function                     |
//| that compares two bars                                           |
//+------------------------------------------------------------------+
enum CPCCACMR
  {
   CPCCACMR_CompareError,              // If two bars with different time are compared
   CPCCACMR_Equal,                     // Bares are equivalent
   CPCCACMR_Different                  // Bars are different
  };

//+------------------------------------------------------------------+
//| Prices that should be used when comparing bars                   |
//+------------------------------------------------------------------+
const uchar CPCH_time = 0;
const uchar CPCH_open = 1;
const uchar CPCH_high = 2;
const uchar CPCH_low = 4;
const uchar CPCH_close = 8;
const uchar CPCH_tick_volume = 16;
const uchar CPCH_spread = 32;
const uchar CPCH_real_volume = 64;


//+------------------------------------------------------------------+
//|  CCustPrevCalculated for calculation of prev_calculated     |
//+------------------------------------------------------------------+
class CCustPrevCalculated
  {
protected:
   // Main parameters
   bool Initialized;                         // The variable for protection from a programming error of class initialization
   string Instrument;                        // stores symbol
   ENUM_TIMEFRAMES TimeFrame;                // stores period
   int BarsLimit;                            // Limit of bars for calculation is set once during initialization of the class
   // Synchronization of history
   uchar HistorySynch;                       // Which exact prices/volumes/spreads should be synchronized
   CPCHistorySynchMode HistorySynchMode;     // Controlling the history from changes by the server post factum when losing connection
   uint HistorySynchSecond;                  // Amount of time (in seconds) back, when the history of DC might be synchronized, 0 - the entire history
   // Array for storing data used for calculation of indicators, allocated once
   uint DataBarsCount;                       // The number of significant bars for that are stored in the array for calculation, cannot exceed BarsLimit
   uint DataBarsCalculated;                  // The number of bars that are not changed for sure
   uint DataStartInd;                        // Initial value of indexing in the array for calculation (for wraparound addressing)
   MqlRates Data[];                          // Array of data for the calculation itself Attention! Wraparound addressing!
   //   
   bool DataBarsCuttingLeft;                 // Set at every call of the PrepareData() procedure at cutting the history from the left
private:
   // Auxiliary variables
   //----------------------------
   MqlRates tmpRates[];                                                                      // Temporary buffer for storying copied data
   CPCCACMR _CompareAndChangeMqlRates(MqlRates &MqlRatesTarget, MqlRates &MqlRatesSource);   // Comparing two variables with the filter HistorySynch
   int _CompareHistory(int LastBarInd, datetime TimeCompare, bool &FullInitialization);      // Comparing the history starting from a specified bar
   // Physical address of a cell according to its virtual number
   uint PInd(uint I, bool AsSeries) {return( AsSeries?((DataStartInd+DataBarsCount-1-I) % BarsLimit):((DataStartInd+I) % BarsLimit) );}
public:
   // Constructor
   void CCustPrevCalculated() {Initialized = false;}
   bool ItIsInitialized() {return(Initialized);}
   // Initialization of operation of the class (executed manually once after creating the class)
   bool InitData(string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute);
   // Main procedure of synchronization of data
   CPCPrepareDataResultCode PrepareData();
   // Functions of access to the synchronized data
   uint GetDataBarsCount() {return(DataBarsCount);}                              // Returns the number of available bars
   uint GetDataBarsCalculated() {return(DataBarsCalculated);}                    // Returns the number of unchanged bars
   uint GetDataStartInd() {return(DataStartInd);}                                // Returns the index for the wraparound acces
   bool GetDataBarsCuttingLeft() {return(DataBarsCuttingLeft);}                  // Returns the result of cutting of bars from the left
      // Note: these functions doesn't control exceeding of a bar's index over the total number
   double GetDataOpen(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].open:Data[(DataStartInd+shift) % BarsLimit].open);}
   double GetDataHigh(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].high:Data[(DataStartInd+shift) % BarsLimit].high);}
   double GetDataLow(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].low:Data[(DataStartInd+shift) % BarsLimit].low);}
   double GetDataClose(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].close:Data[(DataStartInd+shift) % BarsLimit].close);}
   datetime GetDataTime(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].time:Data[(DataStartInd+shift) % BarsLimit].time);}
   long GetDataTick_volume(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].tick_volume:Data[(DataStartInd+shift) % BarsLimit].tick_volume);}
   long GetDataReal_volume(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].real_volume:Data[(DataStartInd+shift) % BarsLimit].real_volume);}
   int GetDataSpread(int shift, bool AsSeries) {return(AsSeries?Data[(DataStartInd+DataBarsCount-1-shift) % BarsLimit].spread:Data[(DataStartInd+shift) % BarsLimit].spread);}
  };
//+------------------------------------------------------------------+
bool CCustPrevCalculated::InitData(string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute)
{
   int tmpLimit = Limit;
   if (tmpLimit > TerminalInfoInteger(TERMINAL_MAXBARS))
      tmpLimit = TerminalInfoInteger(TERMINAL_MAXBARS);
   if (tmpLimit < 2)
      tmpLimit = 2;
   
   CPCHistorySynchMode tmpHSM = HSM;
   if (MQL5InfoInteger(MQL5_TESTING))
      tmpHSM = CPCHSM_NotSynch;
   
   uchar tmpHS = HS;
   if (tmpHS == 0)
      tmpHS = CPCH_open|CPCH_high|CPCH_low|CPCH_close|CPCH_tick_volume|CPCH_spread|CPCH_real_volume;
   
   uint tmpHSMinute = HSMinute * 60;
   
   if (!Initialized)
     {
      BarsLimit = tmpLimit;
      if (ArrayResize(Data, BarsLimit) == -1)
         return(false);
      
      Instrument = Instr;
      TimeFrame = TF;
      HistorySynchMode = tmpHSM;
      HistorySynchSecond = tmpHSMinute;
      
      DataBarsCount = 0; // Start of calculations will be determined by this variable
      
      HistorySynch = tmpHS;
      
      Initialized = true;
      return(true);
     } else {
      // The class is initialized, check the correctness of parameters
      bool res = true;
      if (Instrument != Instr) res = false;
      if (TimeFrame != TF) res = false;
      if (BarsLimit != tmpLimit) res = false;
      if (HistorySynchMode != tmpHSM) res = false;
      if (HistorySynch != tmpHS) res = false;
      if (HistorySynchSecond != tmpHSMinute) res = false;
      return(res);
     }
}
//+------------------------------------------------------------------+
CPCCACMR CCustPrevCalculated::_CompareAndChangeMqlRates(MqlRates &MqlRatesTarget, MqlRates &MqlRatesSource)
{
   if (MqlRatesTarget.time != MqlRatesSource.time)
      return(CPCCACMR_CompareError);
   
   CPCCACMR res = CPCCACMR_Equal;
   
   if ( (HistorySynch & CPCH_open) != 0)
      if (MqlRatesTarget.open != MqlRatesSource.open)
        { MqlRatesTarget.open = MqlRatesSource.open; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_high) != 0)
      if (MqlRatesTarget.high != MqlRatesSource.high)
        { MqlRatesTarget.high = MqlRatesSource.high; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_low) != 0)
      if (MqlRatesTarget.low != MqlRatesSource.low)
        { MqlRatesTarget.low = MqlRatesSource.low; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_close) != 0)
      if (MqlRatesTarget.close != MqlRatesSource.close)
        { MqlRatesTarget.close = MqlRatesSource.close; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_tick_volume) != 0)
      if (MqlRatesTarget.tick_volume != MqlRatesSource.tick_volume)
        { MqlRatesTarget.tick_volume = MqlRatesSource.tick_volume; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_spread) != 0)
      if (MqlRatesTarget.spread != MqlRatesSource.spread)
        { MqlRatesTarget.spread = MqlRatesSource.spread; res = CPCCACMR_Different; }
   
   if ( (HistorySynch & CPCH_real_volume) != 0)
      if (MqlRatesTarget.real_volume != MqlRatesSource.real_volume)
        { MqlRatesTarget.real_volume = MqlRatesSource.real_volume; res = CPCCACMR_Different; }

   return(res);
}
//+------------------------------------------------------------------+
int CCustPrevCalculated::_CompareHistory(int LastBarInd, datetime TimeCompare, bool &FullInitialization)
{
   int res = -1;
   uint tmpHSS = HistorySynchSecond;
   for (int i = LastBarInd; i >= 0; i--)
     {
      // If a limit on the last number of minutes
      // of history synchronization is set, break the cycle
      if (tmpHSS != 0)
         if ( (tmpRates[i].time + HistorySynchSecond) < TimeCompare)
            break;
      switch(_CompareAndChangeMqlRates(Data[PInd(i, false)], tmpRates[i]))
        {
         case CPCCACMR_Different:
            res = (res == -1)?i:MathMin(res, i);
            tmpHSS = 0; // If a difference is found, check the entire history
            break;
         case CPCCACMR_CompareError:
            FullInitialization = true;
            return(-1);
            break;
        }
     }
   return(res);
}
//+------------------------------------------------------------------+
CPCPrepareDataResultCode CCustPrevCalculated::PrepareData()
{
   if (!Initialized)
      return(CPCPDRC_NoData);
      
   int i;
   int BarsCopy = CopyRates(Instrument, TimeFrame, 0, BarsLimit, tmpRates);
   if (BarsCopy <= 0)
      return(CPCPDRC_NoData);
   // Now the tmpRates array contains updated data in the quantity of BarsCopy bars
   // This value can be lower than it was before as a result of hypothetical 
   // cutting of the history by the server
   // the index 0 corresponds to the start of history, the index (BarsCopy-1) corresponds to the last bar
   
   // Prepare auxiliary variables
   datetime DataFirstBarTime = 0;                              // Start date
   datetime DataLastBarTime = 0;                               // End date
   datetime tmpRatesFirstBarTime = tmpRates[0].time;           // Start date
   datetime tmpRatesLastBarTime = tmpRates[BarsCopy-1].time;   // End date
   
   // Flag of full initialization of history as a result of losing connection, synchronizations, etc.
   bool FullInitialization = false;
   
   // If DataBarsCount = 0, then the initialization of history is primary
   bool tmpInsertBars = false;
   if (DataBarsCount != 0)
     {
      DataFirstBarTime = Data[PInd(0, false)].time;              // Start date
      DataLastBarTime = Data[PInd(DataBarsCount-1, false)].time; // End date
      // Checking if new bars are added to previous calculation by the number of bars.
      int BarsOldTest = Bars(Instrument, TimeFrame, DataFirstBarTime, DataLastBarTime);
      if (BarsOldTest == 0)
         return(CPCPDRC_NoData);
      if (BarsOldTest == DataBarsCount)
        {
         if (tmpRatesFirstBarTime >= DataLastBarTime)  // too new history
            FullInitialization = true;
          if (tmpRatesFirstBarTime < DataFirstBarTime) // the history expanded too much t the left
            FullInitialization = true;
        } else {
         // New bars added to the calculated period
         tmpInsertBars = true;
         //FullInitialization = true; // Leave as it is
        }
     } else
      FullInitialization = true;
   
   // Find the index DataLastBarTime in the temporary array
   int SearchIndLB = -1;
   if (!FullInitialization)
     {
      for (i = (BarsCopy-1); i >= 0; i--)
         if (tmpRates[i].time == DataLastBarTime)
           { SearchIndLB = i; break; }
      if (SearchIndLB == -1)
         FullInitialization = true;
     }
   
   // Find the index tmpRatesFirstBarTime in the data array
   int SearchIndRtime0 = -1; // Actually this index means the number of bars that should be cut
   if ( (!FullInitialization) && (tmpRatesFirstBarTime > DataFirstBarTime) ) // The history has been moved as it has been cut from the left
     {
      for (i = 0; i < (int)DataBarsCount; i++)
        {
         if (Data[PInd(i, false)].time == tmpRatesFirstBarTime)
           { SearchIndRtime0 = i; break; }
        }
      if (SearchIndRtime0 == -1)
         FullInitialization = true;
     }
   
   CPCPrepareDataResultCode tmpCPCPDRC = CPCPDRC_Synch;
   if (!FullInitialization)
     {
      // The history has been moved as it has been cut from the left
      DataBarsCuttingLeft = false;
      if (SearchIndRtime0 != -1)
        { // This action means that unwanted bars should be cut from the left side
         DataBarsCount = DataBarsCount - SearchIndRtime0;
         DataStartInd = PInd(SearchIndRtime0, false);
         DataFirstBarTime = tmpRatesFirstBarTime;
         DataBarsCuttingLeft = true;
        }
      
      if (tmpInsertBars)
        {
         int SearchIndInsBars = -1;
         for (i = 0; i < (int)DataBarsCount; i++)
            if (Data[PInd(i, false)].time != tmpRates[i].time)
              {
               SearchIndInsBars = i;
               break;
              }
         if (SearchIndInsBars > 0)
           {
            DataBarsCount = SearchIndInsBars;
            SearchIndLB = SearchIndInsBars - 1;
           } else {
            FullInitialization = true;
           }
        }
     }
     
   DataBarsCalculated = SearchIndLB;
   // It is left to add new bars and shift the indexation!
   if (!FullInitialization)
     {
      uint NewBars = BarsCopy-SearchIndLB-1;
      int tmpCH;
      if (NewBars > 0)
        {
         // The last bar DataLastBarTime will be re-synchronized in any case
         uint BarsToCopy = NewBars + 1;
         DataBarsCount = DataBarsCount - 1;
         
         // Stick new history in this block
         uint RightFreeBars = (BarsLimit - 1) - PInd(DataBarsCount - 1, false);
         DataBarsCount = DataBarsCount + BarsToCopy;
         uint ToCopyRight = MathMin(BarsToCopy, RightFreeBars);
         if (RightFreeBars != 0)
           {
            ArrayCopy(Data, tmpRates, BarsLimit - RightFreeBars, SearchIndLB, ToCopyRight);
            BarsToCopy = BarsToCopy - ToCopyRight;
           } else
            ToCopyRight = 0;
         if (BarsToCopy > 0)
            ArrayCopy(Data, tmpRates, 0, SearchIndLB + ToCopyRight, BarsToCopy);
         DataLastBarTime = tmpRatesLastBarTime;
         
         // Synchronization of history (without new bar)
         if (HistorySynchMode != CPCHSM_NotSynch)
            if (SearchIndLB >= 1)
              {
               tmpCH = _CompareHistory(SearchIndLB - 1, TimeCurrent() - (NewBars * PeriodSeconds(TimeFrame)), FullInitialization);
               if (tmpCH != -1)
                  DataBarsCalculated = (tmpCH==0)?0:tmpCH - 1;
              }
        } else {
         // New bar hasn't been added
         //--------------------------
         // Possibly, the last bar has been recalculated
         switch(_CompareAndChangeMqlRates(Data[PInd(DataBarsCount - 1, false)], tmpRates[BarsCopy-1]))
           {
            case CPCCACMR_Equal:
               DataBarsCalculated = DataBarsCalculated + 1;
               tmpCPCPDRC = CPCPDRC_NoRecountNotRequired;
               break;
            case CPCCACMR_Different:
               tmpCPCPDRC = CPCPDRC_SynchOnlyLastBar;
               break;
            case CPCCACMR_CompareError:
               FullInitialization = true;
               break;
           }
         
         if (!FullInitialization)
           {
            // Synchronization of history (without new bar)
            if (HistorySynchMode == CPCHSM_Paranoid)
               if (SearchIndLB >= 1)
                 {
                  tmpCH = _CompareHistory(SearchIndLB - 1, TimeCurrent(), FullInitialization);
                  if (tmpCH != -1)
                    {
                     DataBarsCalculated = (tmpCH==0)?0:tmpCH - 1;
                     tmpCPCPDRC = CPCPDRC_Synch;
                    }
                 }
           }
        } // if (NewBars > 0)
     }

   if (FullInitialization)
     {
      DataBarsCount = BarsCopy;
      DataStartInd = 0;
      DataBarsCalculated = 0;
      ArrayCopy(Data, tmpRates, 0, 0, DataBarsCount);
      tmpCPCPDRC = CPCPDRC_FullInitialization;
     }
   return(tmpCPCPDRC);
}